import sys
import os
import time
import io
import json
from functools import partial

if sys.version_info >= (3, 10):
    import importlib.resources as ilr
else:
    import importlib_resources as ilr

import urllib
from urllib.error import HTTPError
import unittest
from unittest.mock import patch
import pytest

from fake_useragent import errors, settings, utils
from fake_useragent.utils import urlopen_has_ssl_context
from tests.utils import _request


class TestUtils(unittest.TestCase):
    def setUp(self):
        try:
            os.remove("/tmp/custom.json")
        except OSError:
            pass

    def tearDown(self):
        try:
            os.remove("/tmp/custom.json")
        except OSError:
            pass

    def test_utils_get(self):
        # Good weather test, using local asset as data to be used for mocking
        data = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=data):
            res = utils.get("https://websitesthatdoesexists.com")
            # Res should contain the HTML page now
            self.assertTrue(res)
            self.assertIsInstance(res, str)

    def test_utils_get_retries(self):
        # Raise time-out exception on urlopen method
        with patch.object(urllib.request, "urlopen") as mocked_urlopen:
            mocked_urlopen.side_effect = HTTPError(
                "https://exampe.com", 404, "oopsy", [], []
            )
            with pytest.raises(
                errors.FakeUserAgentError, match="Maximum amount of retries reached"
            ):
                utils.get("https://sitethattimesout.com")

    def test_utils_load(self):
        _load = utils.load

        with patch(
            "fake_useragent.utils.load",
            side_effect=_load,
        ) as mocked:
            browsers = [
                "chrome",
                "edge",
                "internet explorer",
                "firefox",
                "safari",
                "opera",
            ]
            data = utils.load(browsers)

            mocked.assert_called()

        self.assertIsInstance(data["chrome"], list)
        self.assertIsInstance(data["edge"], list)
        self.assertIsInstance(data["firefox"], list)
        self.assertIsInstance(data["opera"], list)
        self.assertIsInstance(data["safari"], list)
        self.assertIsInstance(data["internet explorer"], list)

    def test_utils_write(self):
        path = "/tmp/custom.json"
        data = {"foo": "bar"}

        utils.write(path, data)

        with open(path, encoding="utf-8") as fp:
            expected = json.loads(fp.read())

        assert expected == data

    def test_utils_read(self):
        path = "/tmp/custom.json"
        data = {"foo": "bar"}

        with open(path, mode="w", encoding="utf-8") as fp:
            dumped = json.dumps(data)

            if not isinstance(dumped, utils.text):  # Python 2
                dumped = dumped.decode("utf-8")

            fp.write(dumped)

        expected = utils.read(path)

        assert expected == data

    def test_utils_exist(self):
        path = "/tmp/custom.json"

        assert not os.path.isfile(path)

        assert not utils.exist(path)

        with open(path, mode="wb") as fp:
            fp.write(b"\n")

        assert os.path.isfile(path)

        assert utils.exist(path)

    def test_utils_rm(self):
        path = "/tmp/custom.json"
        assert not os.path.isfile(path)

        with open(path, mode="wb") as fp:
            fp.write(b"\n")

        assert os.path.isfile(path)

        utils.rm(path)

        assert not os.path.isfile(path)

    def test_utils_update(self):
        path = "/tmp/custom.json"

        file = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=file):
            utils.update(path, ["chrome"])

        mtime = os.path.getmtime(path)

        _load = utils.load

        # We need to add a sleep in the unit test (not ideal),
        # otherwise the test could run so fast, we get the same last modification time
        time.sleep(0.1)

        with patch(
            "fake_useragent.utils.load",
            side_effect=_load,
        ) as mocked:
            file = open("tests/assets/chrome.html", "r", encoding="utf-8")
            with patch.object(urllib.request, "urlopen", return_value=file):
                utils.update(path, ["chrome"])

            mocked.assert_called()

        assert os.path.getmtime(path) != mtime

    def test_utils_load_use_local_file(self):
        browsers = [
            "chrome",
            "edge",
            "internet explorer",
            "firefox",
            "safari",
            "opera",
        ]
        # By default use_local_file is also True during production
        data = utils.load(browsers, use_local_file=True)

        self.assertTrue(data["chrome"])
        self.assertTrue(data["edge"])
        self.assertTrue(data["firefox"])
        self.assertTrue(data["opera"])
        self.assertTrue(data["safari"])
        self.assertTrue(data["internet explorer"])
        self.assertIsInstance(data["chrome"], list)
        self.assertIsInstance(data["edge"], list)
        self.assertIsInstance(data["firefox"], list)
        self.assertIsInstance(data["opera"], list)
        self.assertIsInstance(data["safari"], list)
        self.assertIsInstance(data["internet explorer"], list)

    def test_utils_load_use_local_file_pkg_resource_fallback(self):
        browsers = [
            "chrome",
            "edge",
            "internet explorer",
            "firefox",
            "safari",
            "opera",
        ]
        # By default use_local_file is also True during production
        # We will not allow the default importlib resources to be used, by triggering an Exception
        with patch.object(ilr, "files") as mocked_importlib_resources_files:
            # This exception should trigger the alternative path, trying to use pkg_resource as fallback
            mocked_importlib_resources_files.side_effect = Exception("Error")
            data = utils.load(browsers, use_local_file=True)

        self.assertTrue(data["chrome"])
        self.assertTrue(data["edge"])
        self.assertTrue(data["firefox"])
        self.assertTrue(data["opera"])
        self.assertTrue(data["safari"])
        self.assertTrue(data["internet explorer"])
        self.assertIsInstance(data["chrome"], list)
        self.assertIsInstance(data["edge"], list)
        self.assertIsInstance(data["firefox"], list)
        self.assertIsInstance(data["opera"], list)
        self.assertIsInstance(data["safari"], list)
        self.assertIsInstance(data["internet explorer"], list)

    def test_utils_load_cached(self):
        path = "/tmp/custom.json"
        _load = utils.load

        with patch(
            "fake_useragent.utils.load",
            side_effect=_load,
        ) as mocked:
            file = open("tests/assets/chrome.html", "r", encoding="utf-8")
            with patch.object(urllib.request, "urlopen", return_value=file):
                data = utils.load_cached(path, ["chrome"])

            mocked.assert_called()

        self.assertTrue(data["chrome"])
        self.assertIsInstance(data["chrome"], list)

        data = []
        with patch("fake_useragent.utils.load") as mocked:
            file = open("tests/assets/chrome.html", "r", encoding="utf-8")
            with patch.object(urllib.request, "urlopen", return_value=file):
                data = utils.load_cached(path, ["chrome"])

            mocked.assert_not_called()

        self.assertTrue(data["chrome"])
        self.assertIsInstance(data["chrome"], list)

    def test_utils_load_no_local_file_external_data_bad_weather(self):
        path = "/tmp/custom.json"
        denied_urls = [
            "https://useragentstring.com",
        ]

        with patch.object(
            urllib.request,
            "Request",
            side_effect=partial(_request, denied_urls=denied_urls),
        ):
            browsers = [
                "chrome",
                "edge",
                "internet explorer",
                "firefox",
                "safari",
                "opera",
            ]
            with pytest.raises(errors.FakeUserAgentError):
                utils.load(browsers, use_local_file=False)

            with pytest.raises(errors.FakeUserAgentError):
                utils.load_cached(path, browsers)

            with pytest.raises(errors.FakeUserAgentError):
                utils.update(path, browsers)

    def test_utils_load_use_not_local_file_external_is_down(self):
        denied_urls = [
            "https://useragentstring.com/",
        ]

        with patch.object(
            urllib.request,
            "Request",
            side_effect=partial(_request, denied_urls=denied_urls),
        ):
            browsers = [
                "chrome",
                "edge",
                "internet explorer",
                "firefox",
                "safari",
                "opera",
            ]
            with pytest.raises(errors.FakeUserAgentError):
                utils.load(browsers, use_local_file=False)



from fake_useragent.fake import FakeUserAgent, UserAgent  # noqa # isort:skip
from fake_useragent.errors import (
    FakeUserAgentError,
    UserAgentError,
)  # noqa # isort:skip
from fake_useragent.settings import __version__ as VERSION  # noqa # isort:skip



import random
from threading import Lock

from fake_useragent import settings
from fake_useragent.errors import FakeUserAgentError
from fake_useragent.log import logger
from fake_useragent.utils import load, load_cached, str_types, update


class FakeUserAgent:
    def __init__(
        self,
        use_external_data=False,
        cache_path=settings.DB,
        fallback=None,
        browsers=["chrome", "edge", "internet explorer", "firefox", "safari", "opera"],
        verify_ssl=True,
        safe_attrs=tuple(),
    ):
        assert isinstance(
            use_external_data, bool
        ), "use_external_data must be True or False"

        self.use_external_data = use_external_data

        assert isinstance(cache_path, str_types), "cache_path must be string or unicode"

        self.cache_path = cache_path

        if fallback is not None:
            assert isinstance(fallback, str_types), "fallback must be string or unicode"

        self.fallback = fallback

        assert isinstance(browsers, (list, str)), "browsers must be list or string"

        self.browsers = browsers

        assert isinstance(verify_ssl, bool), "verify_ssl must be True or False"

        self.verify_ssl = verify_ssl

        assert isinstance(
            safe_attrs, (list, set, tuple)
        ), "safe_attrs must be list\\tuple\\set of strings or unicode"

        if safe_attrs:
            str_types_safe_attrs = [isinstance(attr, str_types) for attr in safe_attrs]

            assert all(
                str_types_safe_attrs
            ), "safe_attrs must be list\\tuple\\set of strings or unicode"

        self.safe_attrs = set(safe_attrs)

        # initial empty data
        self.data_browsers = {}

        self.load()

    def load(self):
        try:
            with self.load.lock:
                if self.use_external_data:
                    # Use external resource to retrieve browser data
                    self.data_browsers = load_cached(
                        self.cache_path,
                        self.browsers,
                        verify_ssl=self.verify_ssl,
                    )
                else:
                    # By default we will try to load our local file
                    self.data_browsers = load(
                        self.browsers,
                        verify_ssl=self.verify_ssl,
                    )
        except FakeUserAgentError:
            if self.fallback is None:
                raise
            else:
                logger.warning(
                    "Error occurred during fetching data, "
                    "but was suppressed with fallback.",
                )

    load.lock = Lock()

    def update(self, use_external_data=None):
        with self.update.lock:
            if use_external_data is not None:
                assert isinstance(
                    use_external_data, bool
                ), "use_external_data must be True or False"

                self.use_external_data = use_external_data

            # Update tmp cache file from external data source
            if self.use_external_data:
                update(
                    self.cache_path,
                    self.browsers,
                    verify_ssl=self.verify_ssl,
                )

            self.load()

    update.lock = Lock()

    def __getitem__(self, attr):
        return self.__getattr__(attr)

    def __getattr__(self, attr):
        if attr in self.safe_attrs:
            return super(UserAgent, self).__getattr__(attr)

        try:
            for value, replacement in settings.REPLACEMENTS.items():
                attr = attr.replace(value, replacement)

            attr = attr.lower()

            if attr == "random":
                # Pick a random browser from the browsers argument list
                browser_name = random.choice(self.browsers)
            else:
                browser_name = settings.SHORTCUTS.get(attr, attr)

            # Pick a random user-agent string for a specific browser
            return random.choice(self.data_browsers[browser_name])
        except (KeyError, IndexError):
            if self.fallback is None:
                raise FakeUserAgentError(
                    f"Error occurred during getting browser: {attr}"
                )  # noqa
            else:
                logger.warning(
                    f"Error occurred during getting browser: {attr}, "
                    "but was suppressed with fallback.",
                )

                return self.fallback

    @property
    def chrome(self):
        return self.__getattr__("chrome")

    @property
    def googlechrome(self):
        return self.chrome

    @property
    def edge(self):
        return self.__getattr__("edge")

    @property
    def ie(self):
        return self.__getattr__("ie")

    @property
    def internetexplorer(self):
        return self.ie

    @property
    def msie(self):
        return self.ie

    @property
    def firefox(self):
        return self.__getattr__("firefox")

    @property
    def ff(self):
        return self.firefox

    @property
    def safari(self):
        return self.__getattr__("safari")

    @property
    def opera(self):
        return self.__getattr__("opera")

    @property
    def random(self):
        return self.__getattr__("random")


# common alias
UserAgent = FakeUserAgent



import os
import tempfile

try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata

__version__ = metadata.version("fake-useragent")

DB = os.path.join(
    tempfile.gettempdir(),
    "fake_useragent_{version}.json".format(
        version=__version__,
    ),
)

BROWSER_BASE_PAGE = "https://useragentstring.com/pages/{browser}/"  # noqa

BROWSERS_COUNT_LIMIT = 50

REPLACEMENTS = {
    " ": "",
    "_": "",
}

SHORTCUTS = {
    "internetexplorer": "internet explorer",
    "ie": "internet explorer",
    "msie": "internet explorer",
    "microsoft edge": "edge",
    "google": "chrome",
    "googlechrome": "chrome",
    "ff": "firefox",
}

OVERRIDES = {
    "Edge/IE": "Internet Explorer",
    "IE/Edge": "Internet Explorer",
}

HTTP_TIMEOUT = 5

HTTP_RETRIES = 2

HTTP_DELAY = 0.1



import os
import urllib
from urllib import request
from functools import partial

import unittest
from unittest.mock import patch
import pytest

from fake_useragent import (
    VERSION,
    FakeUserAgent,
    FakeUserAgentError,
    UserAgent,
    settings,
)
from tests.utils import _request


class TestFake(unittest.TestCase):
    def setUp(self):
        try:
            os.remove(settings.DB)
        except OSError:
            pass

        try:
            os.remove("/tmp/custom.json")
        except OSError:
            pass

    def tearDown(self):
        try:
            os.remove(settings.DB)
        except OSError:
            pass

        try:
            os.remove("/tmp/custom.json")
        except OSError:
            pass

    def _probe(self, ua):
        ua.ie
        ua.msie
        ua.internetexplorer
        ua.internet_explorer
        ua["internet explorer"]
        ua.edge
        ua.google
        ua.chrome
        ua.googlechrome
        ua.google_chrome
        ua["google chrome"]
        ua.firefox
        ua.ff
        ua.safari
        ua.random
        ua["random"]
        ua.edge

    def test_fake_init(self):
        ua = UserAgent()

        self.assertTrue(ua.chrome)
        self.assertIsInstance(ua.chrome, str)
        self.assertTrue(ua.edge)
        self.assertIsInstance(ua.edge, str)
        self.assertTrue(ua["internet explorer"])
        self.assertIsInstance(ua["internet explorer"], str)
        self.assertTrue(ua.firefox)
        self.assertIsInstance(ua.firefox, str)
        self.assertTrue(ua.safari)
        self.assertIsInstance(ua.safari, str)
        self.assertTrue(ua.opera)
        self.assertIsInstance(ua.opera, str)
        self.assertTrue(ua.random)
        self.assertIsInstance(ua.random, str)

    def test_fake_user_agent_browsers(self):
        ua = UserAgent()

        self._probe(ua)

        with pytest.raises(FakeUserAgentError):
            ua.non_existing

        with pytest.raises(FakeUserAgentError):
            ua["non_existing"]

        data1 = ua.data_browsers

        # Will no do much by default
        ua.update()

        data2 = ua.data_browsers

        assert data1 == data2

        assert data1 is not data2

    def test_fake_default_path(self):
        assert not os.path.isfile(settings.DB)

        data = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=data):
            ua = UserAgent(use_external_data=True, browsers=["chrome"])

        assert settings.DB == ua.cache_path

        assert os.path.isfile(settings.DB)

    def test_fake_fallback(self):
        fallback = "Foo Browser"

        denied_urls = [
            "https://useragentstring.com",
        ]

        with patch.object(
            request,
            "Request",
            side_effect=partial(_request, denied_urls=denied_urls),
        ):
            # This should trigger an error internally (since the site url is denied), causing to call the fallback string
            ua = UserAgent(use_external_data=True, fallback=fallback)

        assert ua.random == fallback

        assert ua.ie == fallback

    def test_fake_no_fallback(self):
        denied_urls = [
            "https://useragentstring.com",
        ]

        with patch.object(
            request,
            "Request",
            side_effect=partial(_request, denied_urls=denied_urls),
        ):
            # Since the website is denied and we didn't specify a fallback string,
            # this should raise an error
            with pytest.raises(FakeUserAgentError):
                UserAgent(use_external_data=True)

    def test_fake_update_external(self):
        data = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=data):
            ua = UserAgent(use_external_data=True, browsers=["chrome"])

        # Open the chrome.html file for 2nd time, for the update()
        data = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=data):
            ua.update()

        self.assertTrue(ua.chrome)
        self.assertIsInstance(ua.chrome, str)

    def test_fake_check_cache_path_external_data(self):
        custom_path = "/tmp/custom.json"
        assert not os.path.isfile(custom_path)

        data = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=data):
            ua = UserAgent(
                cache_path=custom_path, use_external_data=True, browsers=["chrome"]
            )

            assert custom_path == ua.cache_path

        assert os.path.isfile(custom_path)

    def test_fake_update_external_chrome_only(self):
        data = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=data):
            ua = UserAgent(use_external_data=True, browsers=["chrome"])

        # Open the chrome.html file for 2nd time, for the update()
        data = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=data):
            ua.update()

        # Check not being empty
        self.assertTrue(ua.chrome)
        # Is string
        self.assertIsInstance(ua.chrome, str)

    def test_fake_update_external_data_cache(self):
        custom_path = "/tmp/custom.json"
        assert not os.path.isfile(custom_path)

        data = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=data):
            ua = UserAgent(
                cache_path=custom_path, use_external_data=True, browsers=["chrome"]
            )

        with pytest.raises(AssertionError):
            ua.update(use_external_data="y")

        # Open the chrome.html file for 2nd time, for the update()
        data = open("tests/assets/chrome.html", "r", encoding="utf-8")
        with patch.object(urllib.request, "urlopen", return_value=data):
            ua.update(use_external_data=True)

        assert os.path.isfile(custom_path)

        self.assertTrue(ua.chrome)
        self.assertIsInstance(ua.chrome, str)
        self.assertTrue(ua.random)
        self.assertIsInstance(ua.random, str)

    def test_fake_update_external_data_cache_bad_weather(self):
        denied_urls = [
            "https://useragentstring.com",
        ]

        with patch.object(
            request,
            "Request",
            side_effect=partial(_request, denied_urls=denied_urls),
        ):
            with pytest.raises(FakeUserAgentError):
                ua = UserAgent(use_external_data=True)

    def test_fake_external_data_boolean(self):
        with pytest.raises(AssertionError):
            UserAgent(use_external_data="y")

    def test_fake_path_str_types(self):
        with pytest.raises(AssertionError):
            UserAgent(cache_path=10)

    def test_fake_fallback_str_types(self):
        with pytest.raises(AssertionError):
            UserAgent(fallback=True)

    def test_fake_safe_attrs_iterable_str_types(self):
        with pytest.raises(AssertionError):
            UserAgent(safe_attrs={})

        with pytest.raises(AssertionError):
            UserAgent(safe_attrs=[66])

    def test_fake_safe_attrs(self):
        ua = UserAgent(safe_attrs=("__injections__",))

        with pytest.raises(AttributeError):
            ua.__injections__

    def test_fake_version(self):
        assert VERSION == settings.__version__

    def test_fake_aliases(self):
        assert FakeUserAgent is UserAgent



import sys
import contextlib
import inspect
import io
import json
import os
import re
import ssl
import time


# We need files() from Python 3.10 or higher
if sys.version_info >= (3, 10):
    import importlib.resources as ilr
else:
    import importlib_resources as ilr

from urllib.error import URLError
from urllib.parse import quote_plus
from urllib import request
from fake_useragent.log import logger

# Fallback method for retrieving data file
try:
    from pkg_resources import resource_filename
except:
    pass

str_types = (str,)
text = str
urlopen_args = inspect.getfullargspec(request.urlopen).kwonlyargs
urlopen_has_ssl_context = "context" in urlopen_args


def get(url, verify_ssl=True):
    attempt = 0

    while True:
        requestObj = request.Request(url)

        attempt += 1

        try:
            if urlopen_has_ssl_context:
                if not verify_ssl:
                    context = ssl._create_unverified_context()
                else:
                    context = None

                with contextlib.closing(
                    request.urlopen(
                        requestObj,
                        timeout=settings.HTTP_TIMEOUT,
                        context=context,
                    )
                ) as response:
                    return response.read()
            else:  # ssl context is not supported ;(
                with contextlib.closing(
                    request.urlopen(
                        requestObj,
                        timeout=settings.HTTP_TIMEOUT,
                    )
                ) as response:
                    return response.read()
        except (URLError, OSError) as exc:
            logger.debug(
                "Error occurred during fetching %s",
                url,
                exc_info=exc,
            )

            if attempt == settings.HTTP_RETRIES:
                raise FakeUserAgentError("Maximum amount of retries reached")
            else:
                logger.debug(
                    "Sleeping for %s seconds",
                    settings.HTTP_DELAY,
                )
                time.sleep(settings.HTTP_DELAY)


def get_browser_user_agents_online(browser, verify_ssl=True):
    """
    Retrieve browser user agent strings from website
    """
    html = get(
        settings.BROWSER_BASE_PAGE.format(browser=quote_plus(browser)),
        verify_ssl=verify_ssl,
    )
    try:
        html = html.decode("utf-8")
    except (UnicodeDecodeError, AttributeError):
        pass
    html = html.split("<div id='liste'>")[1]
    html = html.split("</div>")[0]

    pattern = r"<a href=\'/.*?>(.+?)</a>"
    browsers_iter = re.finditer(pattern, html, re.UNICODE)

    browsers = []

    for browser in browsers_iter:
        if "more" in browser.group(1).lower():
            continue

        browsers.append(browser.group(1))

        if len(browsers) == settings.BROWSERS_COUNT_LIMIT:
            break

    if not browsers:
        raise FakeUserAgentError(
            "No browser user-agent strings found for browser: {browser}".format(
                browser=browser
            )
        )

    return browsers


def load(browsers, use_local_file=True, verify_ssl=True):
    data = {}
    fetch_online = True
    if use_local_file:
        try:
            json_lines = (
                ilr.files("fake_useragent.data").joinpath("browsers.json").read_text()
            )
            for line in json_lines.splitlines():
                data.update(json.loads(line))
            fetch_online = False
            ret = data
        except Exception as exc:
            # Empty data just to be sure
            data = {}
            logger.warning(
                "Unable to find local data/json file or could not parse the contents using importlib-resources. Try pkg-resource next.",
                exc_info=exc,
            )
            try:
                with open(
                    resource_filename("fake_useragent", "data/browsers.json")
                ) as file:
                    json_lines = file.read()
                    for line in json_lines.splitlines():
                        data.update(json.loads(line))
                fetch_online = False
                ret = data
            except Exception as exc2:
                # Empty data just to be sure
                data = {}
                logger.warning(
                    "Could not find local data/json file or could not parse the contents using pkg-resource. Fallback to external resource.",
                    exc_info=exc2,
                )

    # Fallback behaviour or use_external_data parameter is explicitly set to True
    if fetch_online:
        try:
            # For each browser receive the user-agent strings
            for browser_name in browsers:
                browser_name = browser_name.lower().strip()
                data[browser_name] = get_browser_user_agents_online(
                    browser_name,
                    verify_ssl=verify_ssl,
                )
        except Exception:
            raise FakeUserAgentError("Could not load data from external website")
        else:
            ret = data

    if not ret:
        raise FakeUserAgentError("Data dictionary is empty", ret)

    if not isinstance(ret, dict):
        raise FakeUserAgentError("Data is not dictionary ", ret)

    return ret


def write(path, data):
    with open(path, encoding="utf-8", mode="w") as fp:
        dumped = json.dumps(data)

        if not isinstance(dumped, text):  # Python 2
            dumped = dumped.decode("utf-8")

        fp.write(dumped)


def read(path):
    with open(path, encoding="utf-8") as fp:
        return json.loads(fp.read())


def exist(path):
    return os.path.isfile(path)


def rm(path):
    if exist(path):
        os.remove(path)


def update(cache_path, browsers, verify_ssl=True):
    rm(cache_path)

    write(cache_path, load(browsers, use_local_file=False, verify_ssl=verify_ssl))


def load_cached(cache_path, browsers, verify_ssl=True):
    if not exist(cache_path):
        update(cache_path, browsers, verify_ssl=verify_ssl)

    return read(cache_path)


from fake_useragent import settings  # noqa # isort:skip
from fake_useragent.errors import FakeUserAgentError  # noqa # isort:skip
